# !/usr/bin/python
# -*- coding: utf-8 -*-

"""This module provides a class for converting a JSON file to an edge list."""

import sys

from loguru import logger
from tqdm.auto import tqdm

from bibliometa.configuration import BibliometaConfiguration
from bibliometa.config import LOGGING_FORMAT, JSON_EDGELIST_CONVERSION_CONFIG_DEFAULT
from bibliometa.graph.similarity import Similarity
from bibliometa.utils.archives import tar
from bibliometa.utils.utils import MainUtils, DictUtils


class JSON2EdgeList(BibliometaConfiguration):
    """The :class:`~bibliometa.graph.conversion.JSON2EdgeList` allows to configure and run the conversion from an
    input JSON file to an edge list graph representation.

    It extends the abstract :class:`~bibliometa.configuration.BibliometaConfiguration` class.
    """

    def __init__(self, **kwargs):
        """Construct a new :class:`~bibliometa.graph.conversion.JSON2EdgeList`.

        :param kwargs: Arbitrary keyword arguments that are used as configuration keys and values.
            For example, `verbose=True` will make available a configuration key `verbose` with the value
            `True` (i.e., `self.config.verbose` will then return `True`). Configuration can be set during
            initialization as well as after constructing a class instance by calling the `set_config` method
            on a :class:`~bibliometa.graph.conversion.JSON2EdgeList` object.
        """
        super().__init__(JSON_EDGELIST_CONVERSION_CONFIG_DEFAULT, **kwargs)

    def start(self, n=5):
        """Start the conversion.

        :param n: Number that indicates how many elements will be shown in data preview when verbose == True
        :type n: `int`
        :raises FileNotFoundError: If file given in `self.config.i` can not be found.
        """
        # set up logging
        logger.remove()
        logger.add(self.config.log, format=LOGGING_FORMAT, level=self.config.log_level_file)
        if self.config.verbose:
            logger.add(sys.stderr, level=self.config.log_level_std)
        logger.info("Start JSON2EdgeList conversion.")

        # read JSON file
        try:
            data = DictUtils.read_from_json(self.config.i, encoding=self.config.encoding)
        except FileNotFoundError as e:
            raise e

        # Create graph corpus if necessary
        if self.config.create_corpus:
            corpus = GraphCorpus.create(data, self.config)
        else:
            # load graph corpus from existing file
            try:
                filename, suffix, ext = MainUtils.get_file_info(self.config.corpus, suffix=self.config.name)
                corpus = DictUtils.read_from_json(filename + suffix + ext, encoding=self.config.encoding)
            except FileNotFoundError as e:
                raise FileNotFoundError(
                    f"""{e}. You may have to create a corpus first (set 'create_corpus' in configuration to True)."""
                )

        # show n entries from corpus if verbose == True
        if self.config.verbose:
            print(dict(list(corpus.items())[:n]))

        # Calculate similarity
        logger.info("Start similarity calculation.")
        Similarity.calculate(corpus, self.config)
        logger.info("Similarity calculation ended.")

        # put similarity files into tar.gz archive
        if self.config.archive:
            if self.config.archive_ext == ".tar.gz":
                logger.info(f"Putting similarity files into '{self.config.archive_ext}' archives.")
                filename, suffix, ext = MainUtils.get_file_info(self.config.o, suffix=self.config.name)
                tar(
                    filename + suffix + ext,
                    filename + suffix + self.config.archive_ext
                )
                logger.info(f"Similarity '{self.config.archive_ext}' archive can now be found in "
                            f"{filename + suffix + self.config.archive_ext}."
                            )
            else:
                raise ValueError("Archive extensions other than 'tar.gz' are not implemented yet!")


class GraphCorpus:
    """The :class:`~bibliometa.graph.conversion.GraphCorpus` provides a static function to create a graph corpus
    in JSON format. It is needed in the conversion from JSON to an edge list representation.
    """

    @staticmethod
    def create(data, config):
        """Create a graph corpus.

        :param data: Dictionary containing data  sets
        :type data: `dict`
        :param config: Configuration object
        :type config: `bibliometa.configuration.Config`
        :return: Graph corpus
        :rtype: `dict`
        :raise FileNotFoundError: If graph corpus can not be written to file
        """

        def _create_swap():
            """Create a corpus where unique values from the input data become the keys
            and keys from the input data become the values.

            :return: Graph corpus
            :rtype: `dict`
            """
            _corpus = {}
            _unique_values = []

            # Collect unique data values as corpus keys,
            # consider only those fields that are defined in the configuration
            for key in data.keys():
                for field, subfield in config.fields:
                    if field in data[key].keys():
                        if subfield in data[key][field].keys():
                            values = data[key][field][subfield]
                            for value in values:
                                if len(value.strip()) > 0:
                                    _unique_values.append(value.strip())
            _unique_values = list(set(_unique_values))

            # Fill graph corpus
            with tqdm(total=len(_unique_values)) as progressbar:
                for value in _unique_values:
                    _corpus[value] = []
                    for key in data.keys():
                        for field, subfield in config.fields:
                            if field in data[key].keys():
                                if subfield in data[key][field].keys():
                                    sf_values = data[key][field][subfield]
                                    for sf_value in sf_values:
                                        if sf_value.strip() == value:
                                            _corpus[value].append(key)
                                            _corpus[value] = list(set(_corpus[value]))
                    progressbar.update()

            logger.debug(f"Corpus keys: {len(_corpus.keys())}")

            return _corpus

        def _create_original():
            """Create a corpus where keys and values from the input data also become keys
            and values in the graph corpus.

            :return: Graph corpus
            :rtype: `dict`
            """
            _corpus = {}
            _unique_values = []

            # use keys from import file as corpus keys
            for key in data.keys():
                _corpus[key] = []
                for field, subfield in config.fields:
                    if field in data[key].keys():
                        if subfield in data[key][field].keys():
                            values = data[key][field][subfield]
                            for value in values:
                                if len(value.strip()) > 0:
                                    _unique_values.append(value.strip())
                                    _corpus[key].append(value.strip())
                                    _corpus[key] = list(set(_corpus[key]))

                # remove empty dictionary items
                if len(_corpus[key]) == 0:
                    del _corpus[key]

            _unique_values = set(_unique_values)
            logger.debug(f"Corpus keys: {len(_corpus.keys())}")
            logger.debug(f"Unique values: {len(_unique_values)}")

            return _corpus

        # class action starts here
        corpus = _create_swap() if config.swap else _create_original()

        # Write corpus to file
        try:
            filename, suffix, ext = MainUtils.get_file_info(config.corpus, suffix=config.name)
            DictUtils.save_to_json(corpus, filename + suffix + ext, encoding=config.encoding)
            logger.info(f"Corpus written to file {filename + suffix + ext}.")
        except FileNotFoundError as e:
            raise e

        return corpus



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""This module provides utility functions for graphs."""

import json
import os

from operator import itemgetter

import networkx as nx
import pandas as pd

from loguru import logger

from bibliometa.utils.utils import MainUtils

_ENCODING = "utf-8"


def load_graph(config, reload=False):
    """Load a graph from a GraphML or similarity file.

    :param config: Configuration object
    :type config: `bibliometa.configuration.Config`
    :param reload: If graph will be loaded directly from similarity file
    :type reload: `bool`
    :return: Graph
    :rtype: `networkx.Graph`
    """
    # get reload configuration parameter if available
    try:
        reload = config.reload
    except Exception:
        pass

    if not reload:
        if config.graphml:
            if config.graphml.endswith(".graphml"):
                _path = config.graphml
            else:
                filename, suffix, ext = MainUtils.get_file_info(
                    config.graphml + config.config_id + ".graphml", suffix=config.name
                )
                _path = filename + ext + suffix
        else:
            raise ValueError("Configuration parameter 'graphml' is not defined!")
        try:
            logger.info(f"Import from .graphml file {_path}.")
            graph = nx.read_graphml(_path)
            logger.info("Import finished.")
        except FileNotFoundError:
            raise FileNotFoundError(f"GraphML file at {_path} can not be found. If using JSON2EdgeList, "
                                    "do you need to set config.reload to True?")
    else:
        logger.info(f"Create graph from information in file {config.i}.")
        graph = read_file(config, config.encoding)
        logger.info("Graph creation finished.")

    return graph


def read_file(config, encoding=_ENCODING):
    """Create graph by reading in a similarity file.

    :param config: Configuration object
    :type config: `bibliometa.configuration.Config`
    :param encoding: File encoding
    :type encoding: `str`
    :return: Graph
    :rtype: `networkx.Graph`
    """
    g = nx.Graph()

    count = 0
    for df in pd.read_csv(config.i, chunksize=config.chunksize, sep=config.csv_sep, index_col=0, low_memory=False,
                          compression='gzip', encoding=encoding, header=0,
                          usecols=[0, 1, 2, config.sim_functions.index(config.sim) + 3],
                          names=['id', 'nodea', 'nodeb', 'sim']):
        count += 1
        logger.info("Import datasets ...")  # (up to {count * config.chunksize / 1000000} million) ...")
        for i in [tuple(itemgetter(0, 1, 2)(x)) for x in df.values]:
            if i[2] > config.t:
                try:
                    g.add_weighted_edges_from([i])
                except Exception as e:
                    logger.warning(
                        f"can not add weighted edge from {[i]} in chunk {count * config.chunksize}, error {e}"
                    )
    logger.info("Import finished!")

    if config.create_graphml:
        if config.graphml.endswith(".graphml"):
            _path = config.graphml
        else:
            filename, suffix, ext = MainUtils.get_file_info(
                config.graphml + config.config_id + ".graphml", suffix=config.name
            )
            _path = filename + ext + suffix

        dirname = os.path.dirname(_path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        logger.info("Saving graph as .graphml file ...")
        nx.write_graphml(g, _path, encoding=encoding)
        logger.info("Saved .graphml file successfully.")

    return g


def save_file(path, results, encoding=_ENCODING):
    """Save results in a text file.

    :param path: Path to output file
    :type path: `str`
    :param results: List containing results from graph analysis
    :type results: `list`
    :param encoding: Encoding of file
    :type encoding: `str`
    """
    dirname = os.path.dirname(path)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    with open(path, 'w', encoding=encoding) as f:
        f.write("\n".join(results))


def get_nodes(graph, config, encoding=_ENCODING):
    """Get nodes and their degrees from a graph.

    :param graph: Graph object
    :type graph: `networkx.Graph`
    :param config: Configuration object
    :type config: `bibliometa.configuration.Config`
    :param encoding: File encoding
    :type encoding: `str`
    :return: Dictionary of nodes with their degrees
    :rtype: `dict`
    """
    nodes = {n: graph.degree[n] for n in graph.nodes()}
    nodes = dict(sorted(nodes.items(), key=lambda item: item[1], reverse=True))

    _path = f"{config.degrees}degrees_{config.name}.json"
    if not os.path.exists(os.path.dirname(_path)):
        os.makedirs(os.path.dirname(_path))
    with open(_path, "w", encoding=encoding) as f:
        json.dump(nodes, f, indent=4)

    return nodes


def update_graph(graph, df, col):
    """Remove nodes from graph that do not appear in column col in DataFrame df.

    :param graph: Graph object
    :type graph: `networkx.Graph`
    :param df: DataFrame
    :type df: `pandas.DataFrame`
    :param col: Column in DataFrame
    :type col: `str`
    :return: Updated Graph object
    :rtype: `networkx.Graph`
    """
    remove_list = []

    logger.debug(f"Number of nodes in the graph (before update): {len(graph.nodes)}")

    for n in graph.nodes:
        if n not in df[col].tolist():
            remove_list.append(n)

    for n in remove_list:
        try:
            graph.remove_node(n)
        except Exception:
            pass

    logger.debug(f"Number of nodes in the graph (after update): {len(graph.nodes)}")

    for n in graph.nodes:
        logger.debug(n)

    return graph


def get_subgraph(graph):
    """Get largest connected component from graph.

    :param graph: Graph object
    :type graph: `networkx.Graph`
    :return: Largest component
    :rtype: `networkx.Graph`
    """
    subgraphs = [graph.subgraph(c).copy() if graph.subgraph(c) else nx.Graph() for c in nx.connected_components(graph)]
    # TODO: Falls len(subgraphs) > 0: abfangen (keine Übereinstimmung zwischen coordinates[geo_col] und nodes ...)

    maxlen = 0
    largest = subgraphs[0]
    for x in subgraphs:
        if len(x.nodes) > maxlen:
            largest = x
            maxlen = len(x.nodes)

    logger.debug(f"{len(subgraphs)} components found, only use largest component with size {len(largest.nodes)}.")

    return largest


def get_graph_attributes(graph):
    """Get degrees, labels and sizes for a graph.

    :param graph: Graph object
    :type graph: `networkx.Graph`
    :return: Dictionary with degrees, labels, sizes
    :rtype: `dict`
    """
    degrees = nx.degree(graph)
    labels = {n: n if degrees[n] >= 0 else '' for n in graph.nodes}
    sizes = [10 * degrees[n] for n in graph.nodes]

    return {"degrees": degrees,
            "labels": labels,
            "sizes": sizes}


def create_pos(graph, df, geo_col, verbose):
    """Create dictionary with node positions.

    :param graph: Graph object
    :type graph: `networkx.Graph`
    :param df: DataFrame with geographical information
    :type df: `pandas.DataFrame`
    :param geo_col: Column in DataFrame
    :type geo_col: `str`
    :param verbose: Verbose parameter
    :type verbose: `bool`
    :return: Dictionary of positions
    :rtype: `dict`
    """
    pos = {}
    for n in graph.nodes:
        lat = df[df[geo_col] == n]["lat"].to_string(index=False)
        lng = df[df[geo_col] == n]["lng"].to_string(index=False)
        pos[n] = (float(lng), float(lat))

    logger.debug(f"Positions: {pos}")
    if verbose:
        print(pos)

    return pos



# !/usr/bin/python
# -*- coding: utf-8 -*-

"""This module provides classes and functions to visualize network data on geographical maps."""

import os
import sys

import cartopy.crs as ccrs
import cartopy.io.shapereader as shpreader
import geopandas as gpd
import matplotlib.pyplot as plt
import networkx as nx
import pandas as pd

from loguru import logger
from matplotlib.pyplot import text
from shapely.geometry import Point

from bibliometa.config import LOGGING_FORMAT, GRAPH_VISUALISATION_MAP_CONFIG_DEFAULT
from bibliometa.configuration import BibliometaConfiguration
from bibliometa.graph.utils import load_graph, update_graph, get_nodes, get_subgraph, \
    get_graph_attributes, create_pos


class Map(BibliometaConfiguration):
    """The :class:`~bibliometa.graph.visualization.Map` provides functions to visualize geographical network data on
    a map.
    """

    def __init__(self, **kwargs):
        """Construct a new :class:`~bibliometa.graph.visualization.Map`."""
        super().__init__(GRAPH_VISUALISATION_MAP_CONFIG_DEFAULT, **kwargs)

    def _update_config(self):
        """Update configuration with class-specific values and check for configuration correctness."""
        self._shp = MapUtils.read_shp(self.config.shapefile, self.config.shapefile_color, self.config.verbose)
        self._shapes = list(shpreader.Reader(self.config.shapefile).geometries())

        # TODO: not all cities in the CT are represented in this table.
        #  Take the longitude/latitude from the CT instead?
        self._df = MapUtils.convert_to_gdf(self.config.coordinates, self.config.crs, self.config.coordinates_sep)

        self.graph = load_graph(self.config, reload=False)

        # create dictionary of nodes and their node degrees
        self._nodes = get_nodes(self.graph, self.config, encoding=self.config.encoding)

        if not os.path.exists(self.config.o):
            os.makedirs(self.config.o)
        if not os.path.exists(self.config.graphml):
            os.makedirs(self.config.graphml)

    def start(self):
        """Start visualization."""
        # Set up logging
        logger.remove()
        logger.add(self.config.log, format=LOGGING_FORMAT, level=self.config.log_level_file)
        if self.config.verbose:
            logger.add(sys.stderr, level=self.config.log_level_std)
        logger.info("Start network creation.")

        self._update_config()

        # plot all cities from city file as scatter plot
        logger.info("Creating scatterplot.")
        Plotting.scatter(self._df, self.config)

        # plot all cities from city file on map
        logger.info("Plotting all cities.")
        Plotting.cities(self._df, self._shp, self.config)

        # plot nodes on map
        logger.info("Plotting node degrees.")
        Plotting.degrees(self._df, self._shp, self._nodes, self.config)

        # remove unneeded nodes from the graph
        update_graph(self.graph, self._df, self.config.geo_col)

        # get largest component as subgraph and plot on map
        logger.info("Plotting network on map.")
        subgraph = get_subgraph(self.graph)
        Plotting.graph_on_map(self._shapes,
                              subgraph,
                              get_graph_attributes(subgraph),
                              create_pos(self.graph, self._df, self.config.geo_col, self.config.verbose),
                              self.config)

        logger.info("Network creation finished.")
        logger.info(f"Logfile written to {self.config.log}")


class Plotting:
    """The :class:`~bibliometa.graph.visualization.Plotting` provides functions to plot network data."""

    @staticmethod
    def scatter(df, config):
        """Create scatter plot.

        :param df: DataFrame with 'lng' and 'lat' column
        :type df: `pandas.DataFrame`
        :param config: Configuration object
        :type config: `bibliometa.configuration.Config`
        """
        plot = plt.scatter(x=df['lng'], y=df['lat'])
        if config.verbose:
            plt.show()

        for ext in config.o_formats:
            fig = plot.get_figure()
            fig.savefig(f"{config.o}map_scatter_{config.name}.{ext}", bbox_inches='tight')

        plt.clf()

    @staticmethod
    def cities(df, shp, config):
        """Plot cities on a certain map.

        :param df: DataFrame with city information
        :type df: `pandas.DataFrame`
        :param shp: Shapefile
        :type shp: GeoDataFrame
        :param config: Configuration object
        :type config: `bibliometa.configuration.Config`
        """
        plot = df.plot(ax=shp.plot(figsize=config.figsize, marker='o', color=config.shapefile_color, markersize=45))
        if config.verbose:
            plt.show()

        for ext in config.o_formats:
            fig = plot.get_figure()
            fig.savefig(f"{config.o}map_cities_{config.name}.{ext}", bbox_inches='tight')

        plt.clf()

    @staticmethod
    def degrees(df, shp, nodes, config):
        """Plot graph with node degrees.

        :param df: DataFrame with city information
        :type df: `pandas.DataFrame`
        :param nodes: Graph nodes with their degrees
        :type nodes: `dict`
        :param shp: Shapefile
        :type shp: GeoDataFrame
        :param config: Configuration object
        :type config: `bibliometa.configuration.Config`
        """
        fig, ax = plt.subplots(figsize=config.figsize)
        shp.plot(ax=ax, alpha=0.6, color=config.shapefile_color)
        for city in nodes.keys():
            if city in df[config.geo_col].tolist():
                df[df[config.geo_col] == city].plot(ax=ax, markersize=nodes[city],
                                                    color=config.degree_node_color, marker="o")
        for x, y, label in zip(df.geometry.x, df.geometry.y, df.city):
            if label in nodes.keys():
                ax.annotate(label, xy=(x, y), xytext=(3, 3), textcoords="offset points")
        if config.verbose:
            plt.show()

        for ext in config.o_formats:
            fig.savefig(f"{config.o}map_degrees_{config.name}.{ext}", bbox_inches='tight')

        plt.clf()

    @staticmethod
    def graph_on_map(shapes, subgraph, attributes, pos, config):
        """Plot graph on a map.

        :param shapes: An iterator of shapely geometries from a shapefile
        :type shapes: `list`
        :param subgraph: Largest component of graph
        :type subgraph: `networkx.Graph`
        :param attributes: Dictionary of graph degrees, labels and sizes
        :type attributes:  `dict`
        :param pos: Dictionary of node positions
        :type pos: `dict`
        :param config: Configuration object
        :type config: `bibliometa.configuration.Config`
        """
        crs = ccrs.PlateCarree()
        fig, ax = plt.subplots(1, 1, figsize=config.figsize,
                               subplot_kw=dict(projection=crs))

        ax.add_geometries(shapes, crs, edgecolor='black', facecolor='gray', alpha=0.2)

        ax.coastlines()
        # ax.set_global()
        ax.set_extent([5.5, 15.5, 47.2, 55.3])

        nx.draw_networkx(subgraph,
                         ax=ax,
                         # font_size=24,
                         alpha=.5,
                         width=.25,
                         node_size=attributes["sizes"],
                         labels=attributes["labels"],
                         pos=pos,
                         with_labels=False,
                         # node_color=altitude,
                         # cmap=plt.cm.autumn
                         )

        # add labels
        for node, (x, y) in pos.items():
            fs = 12
            try:
                if attributes["degrees"][node] > 12:
                    fs = attributes["degrees"][node]
                text(x, y, node, fontsize=fs, ha='center', va='center')
            except Exception:
                pass

        if config.verbose:
            plt.show()

        for ext in config.o_formats:
            fig.savefig(f"{config.o}map_network_{config.name}.{ext}", bbox_inches='tight')

        plt.clf()


class MapUtils:
    """The :class:`~bibliometa.graph.visualization.MapUtils` provides utilities for the
    :class:`~bibliometa.graph.visualization.Map` class.
     """

    @staticmethod
    def read_shp(f, color="grey", verbose=False):
        """Read shapefile.

        :param f: Path to shapefile
        :type f: `str`
        :param color: Color of shapefile background
        :type color: `str`
        :param verbose: Verbose parameter
        :type verbose: `bool`
        :return: Shapefile
        :rtype: GeoDataFrame
        """
        shp = gpd.read_file(f)
        if verbose:
            fig, ax = plt.subplots(figsize=(15, 15))
            shp.plot(ax=ax, alpha=0.6, color=color)
            plt.show()

        return shp

    @staticmethod
    def convert_to_gdf(csv_input, crs, csv_sep):
        """Convert long/lat to Point objects.

        :param csv_input: Path to CSV input file with long/lat information
        :type csv_input: `str`
        :param crs: Coordinate Reference System
        :type crs: `str`
        :param csv_sep: CSV separator in input file
        :type csv_sep: `str`
        """
        orig_df = pd.read_csv(csv_input, sep=csv_sep)
        df = orig_df.copy()
        try:
            geometry = [Point(xy) for xy in zip(df["lng"], df["lat"])]
            df = gpd.GeoDataFrame(df, crs=crs, geometry=geometry)
            logger.info(f"Successfully converted import data from {csv_input} to GeoDataFrame.")
            logger.debug(f"Imported city data:\n{orig_df.head()}")
            logger.debug(f"GeoDataFrame:\n{df.head()}")
            return df
        except Exception as e:
            raise ValueError(f"Could not convert import data from {csv_input} to GeoDataFrame. Exception: {e}")



# !/usr/bin/python
# -*- coding: utf-8 -*-

"""This module provides the :class:`~bibliometa.conversion.CSV2JSON`."""

import json
import re
import sys

import pandas as pd
from json import JSONDecodeError
from loguru import logger
from tqdm.auto import tqdm

from bibliometa.configuration import BibliometaConfiguration
from bibliometa.config import LOGGING_FORMAT, CSV_JSON_CONVERSION_CONFIG_DEFAULT
from bibliometa.utils.utils import MainUtils, DictUtils


class CSV2JSON(BibliometaConfiguration):
    """The :class:`~bibliometa.conversion.CSV2JSON` allows to configure and run the conversion from an
    input CSV file to a JSON file containing only those information from the CSV file needed for further analysis.

    It extends the abstract :class:`~bibliometa.configuration.BibliometaConfiguration` class.
    """

    def __init__(self, **kwargs):
        """Construct a new :class:`~bibliometa.conversion.CSV2JSON`.

        :param kwargs: Arbitrary keyword arguments that are used as configuration keys and values.
            For example, `verbose=True` will make available a configuration key `verbose` with the value
            `True` (i.e., `self.config.verbose` will then return `True`). Configuration can be set during
            initialization as well as after constructing a class instance by calling the `set_config` method
            on a :class:`~bibliometa.conversion.CSV2JSON` object.
        """
        # TODO: Are the super parameters needed? What do they do?
        # super(CSV2JSON, self).__init__(CSV_JSON_CONVERSION_CONFIG_DEFAULT)
        super().__init__(CSV_JSON_CONVERSION_CONFIG_DEFAULT, **kwargs)

    def _update_config(self):
        """Update configuration with class-specific values and check for configuration correctness.

        :raise ValueError: If `self.config.step` <= 0
        """
        if self.config.step <= 0:
            raise ValueError("Wrong configuration! 'step' must be set to a value > 0.")

    def _save_results(self, _dict, file):
        """Save results to JSON file.

        :param _dict: Dictionary containing results
        :type _dict: `dict`
        :param file: Path to JSON file
        :type file: `str`
        """
        try:
            with open(file, "r", encoding=self.config.encoding) as f:
                try:
                    _d = json.load(f)
                except JSONDecodeError:
                    _d = {}
        except FileNotFoundError:
            _d = {}  # file does not exists in fist iteration, skip
        with open(file, "w", encoding=self.config.encoding) as f:
            # merge existing file and new results
            _dict = DictUtils.merge(_dict, _d)
            json.dump(_dict, f, indent=4)
        return _dict

    def start(self):
        """Start the conversion.

        :raises FileNotFoundError: If file given in `self.config.i` can not be found.
        """

        def _is_in(a, b):
            """Check if b is a value in the range of a.

            :param a: A list of `int` values with either length 1 or 2
            :type a: `list`
            :param b: An integer value, e.g., a year
            :type b: `int`
            :return: `True` if b is in the range of a, `False` otherwise
            :rtype: `bool`

            .. important::
                If parameter `a` is of length 1, this function returns `True` if `b` is either the same as
                `a[0]` or within the interval defined by `self.config.interval_lower` and
                `self.config.interval_upper`. For example, if `a[0]` is 1750 and both `self.config.interval_lower`
                and `self.config.interval_upper` are set to 10, this function returns `True` if `b` has a value
                between 1740 and 1760 (inclusive).
            """
            if len(a) == 1:
                return (b - self.config.interval_lower) <= a[0] <= (b + self.config.interval_upper)
            if len(a) == 2:
                return a[0] <= b <= a[1]
            return False

        def _get_era(d):
            # TODO: Rework this code
            """Extract start and end year from a `str` variable.

            :param d: String with year information
            :type d: `str`
            :return: List of years in `d` or None
            :rtype: `list`, `None`
            """
            # TODO: optimize date parsing (preferably already in input data)
            if self.config.subfield_sep + self.config.datefield[1] in d:
                splitted = d.split(self.config.subfield_sep)

                if splitted[0][4] in self.config.date_indicator:
                    # TODO: handle B.C. dates
                    try:
                        if splitted[3].count("a") > 0 and splitted[3].count("b") == 0:  # ignore B.C. dates
                            # remove first character (i.e., subfield indicator)
                            d_split = re.split('[au]', splitted[3][1:])
                            d_split = [int(x) for x in d_split if len(x.strip()) > 0]
                            return d_split
                        else:
                            return None
                    except Exception:
                        return None

        # set up logging
        logger.remove()
        logger.add(self.config.log, format=LOGGING_FORMAT, level=self.config.log_level_file)
        if self.config.verbose:
            logger.add(sys.stderr, level=self.config.log_level_std)
        logger.info("Start conversion from CSV to JSON.")

        # update configuration
        self._update_config()

        # read CSV file
        try:
            with open(self.config.i, "r", encoding=self.config.encoding) as f:
                df = pd.read_csv(f, sep=self.config.csv_sep)
                df.fillna('', inplace=True)
                logger.info(f"Size of import data: {df.shape[0]} data sets")
        except FileNotFoundError as e:
            raise e

        # calculate size of progress bar
        factor = MainUtils.get_factor(df.shape[0])
        progress_bar_max = int(df.shape[0] / factor)

        # start row-by-row conversion for each defined year
        # TODO: Allow for a conversion that is not based on single years (e.g., using the full input data set)
        # TODO: Log that the conversion process for year XY starts
        year = self.config.from_
        while year <= self.config.to:
            # year is added as suffix to the path from self.config.o
            filename, suffix, ext = MainUtils.get_file_info(self.config.o, suffix=year)
            file = filename + suffix + ext
            with tqdm(total=progress_bar_max) as progressbar:
                _dict = {}  # results dictionary
                for index, row in df.iterrows():
                    row_id = row['id']
                    _dict[row_id] = {}
                    # iterate over all fields defined in configuration
                    for fields in self.config.fields:
                        content_field = fields["content"][0]
                        content_subfield = fields["content"][1]
                        # get dates from row
                        dates = [_get_era(d.strip()) for d in str(row[
                                                                      self.config.datefield[0]
                                                                      + self.config.subfield_sep
                                                                      + self.config.datefield[1]
                                                                      ]).split(self.config.split_char)
                                 if len(d.strip()) > 0 and _get_era(d.strip())]

                        # check if at least one date is in desired period/year
                        # if yes, get content and add to results dictionary where appropriate
                        for date in dates:
                            if _is_in(date, year):
                                contents = [c.strip() for c in str(row[
                                                                       content_field
                                                                       + self.config.subfield_sep
                                                                       + content_subfield
                                                                       ]).split(self.config.split_char)
                                            if len(c.strip()) > 0]
                                types = [t.strip() for t in str(row[
                                                                    fields["type"][0]
                                                                    + self.config.subfield_sep
                                                                    + fields["type"][1]
                                                                    ]).split(self.config.split_char)
                                         if len(t.strip()) > 0]

                                # get only content from desired type(s)
                                content = [c for (c, t) in zip(contents, types) if t in fields["categories"]]

                                if len(content) > 0:
                                    _dict[row_id][content_field] = {content_subfield: content}
                                    break

                    # after each row, remove empty values from dictionary
                    _dict = {k: v for k, v in _dict.items() if len(v.keys()) > 0}

                    # update progress bar and save results dictionary
                    if index % factor == 0 and int(index / factor) != progress_bar_max:
                        progressbar.update()
                        self._save_results(_dict, file)
                        _dict = {}

                # at this point, all rows were processed
                # close progress bar and save final results dictionary as JSON
                progressbar.close()
                _dict = self._save_results(_dict, file)

                logger.info(f"Size of output data for year {year} "
                            f"(+{self.config.interval_lower}/-{self.config.interval_upper} years): "
                            f"{len(_dict.keys())} data sets "
                            f"({round(len(_dict.keys()) * 100 / df.shape[0], 2)} %) "
                            )

                # increase while loop condition
                year = year + self.config.step



# !/usr/bin/python
# -*- coding: utf-8 -*-

"""This module provides functions for analysing graphs."""

import collections
import os
import sys

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from loguru import logger
from tqdm.auto import tqdm

from bibliometa.config import LOGGING_FORMAT, GRAPH_ANALYSIS_CONFIG_DEFAULT
from bibliometa.configuration import BibliometaConfiguration
from bibliometa.graph.utils import load_graph, save_file
from bibliometa.utils.utils import DictUtils, MainUtils


class GraphAnalysis(BibliometaConfiguration):
    """The :class:`~bibliometa.graph.analysis.GraphAnalysis` allows to configure and run the graph analysis of
    a graph corpus.

    It extends the abstract :class:`~bibliometa.configuration.BibliometaConfiguration` class.
    """

    def __init__(self, **kwargs):
        """Construct a new :class:`~bibliometa.graph.analysis.GraphAnalysis`.

        :param **kwargs: Arbitrary keyword arguments that are used as configuration keys and values.
            For example, `verbose=True` will make available a configuration key `verbose` with the value
            `True` (i.e., `self.config.verbose` will then return `True`). Configuration can be set during
            initialization as well as after constructing a class instance by calling the `set_config` method
            on a :class:`~bibliometa.graph.analysis.GraphAnalysis` object.
        """
        super().__init__(GRAPH_ANALYSIS_CONFIG_DEFAULT, **kwargs)

    def _update_config(self):
        """Update configuration with class-specific values and check for configuration correctness.

        :raise ValueError: If no `self.config.sim_functions` are given.
        """
        self.set_config(config_id=f"{self.config.n}_{self.config.e}_{self.config.sim}_{self.config.t}")

        if len(self.config.sim_functions) == 0 or len(self.config.sim) == 0:
            raise ValueError("No similarity function(s) given! At least one similarity function must be provided.")

        try:
            if not os.path.exists(os.path.dirname(self.config.img)):
                os.makedirs(os.path.dirname(self.config.img))
        except Exception:
            pass
        try:
            if not os.path.exists(os.path.dirname(self.config.graphml)):
                os.makedirs(os.path.dirname(self.config.graphml))
        except Exception:
            pass

        if self.config.create_graphml and not self.config.graphml:
            raise ValueError("Configuration parameter 'graphml' must be set if parameter 'create' is set to True,"
                             "otherwise no GraphML file can be created!")

        self._results = []

        # Load graph from GraphML file, if existent; otherwise create GraphML file from similarity file
        self.g = load_graph(self.config)

    def start(self):
        """Start the analysis."""
        # Set up logging
        filename, suffix, ext = MainUtils.get_file_info(
            self.config.log, suffix=self.config.name
        )
        logger.remove()
        logger.add(filename + suffix + ext, format=LOGGING_FORMAT, level=self.config.log_level_file)
        if self.config.verbose:
            logger.add(sys.stderr, level=self.config.log_level_std)

        self._update_config()
        logger.info(f"Start graph analysis for configuration {self.config.config_id}.")

        filename, suffix, ext = MainUtils.get_file_info(
            self.config.o, suffix=self.config.name
        )

        with tqdm(total=9) as progressbar:
            # Calculate basic metrics
            nodes = self._node_count()
            edges = self._edge_count()
            components = self._component_count()
            max_component = self._max_component()
            avg_degree = self._avg_degree()

            # Get basic metrics
            nodes_str = f"Nodes: {nodes}"
            self._results.append(nodes_str)
            logger.info(nodes_str)
            edges_str = f"Edges: {edges}"
            self._results.append(edges_str)
            logger.info(edges_str)
            components_str = f"Components: {components}"
            self._results.append(components_str)
            logger.info(components_str)
            max_component_str = f"Size of largest component: {max_component}"
            self._results.append(max_component_str)
            logger.info(max_component_str)
            avg_degree_str = f"Average Degree: {avg_degree}"
            self._results.append(avg_degree_str)
            logger.info(avg_degree_str)

            save_file(filename + suffix + ext, self._results)
            progressbar.update()

            # Plot degree distribution
            logger.info("Start plotting degree distributions.")
            self._degree_distribution(weight='weight')
            logger.info("Finished plotting degree distributions.")

            # Top degree centrality nodes
            logger.info("Getting top Degree Centrality nodes.")
            self._top_dc_nodes()
            # Degree Centrality distributions
            logger.info("Calculating Degree Centrality distributions.")
            dc_values = self._degree_centrality()
            avg_dc_str = f"Average degree centrality: {np.mean(list(dc_values))}"
            self._results.append(avg_dc_str)
            logger.info(avg_dc_str)
            min_dc_str = f"Minimum degree centrality: {np.min(list(dc_values))}"
            self._results.append(min_dc_str)
            logger.info(min_dc_str)
            max_dc_str = f"Maximum degree centrality: {np.max(list(dc_values))}"
            self._results.append(max_dc_str)
            logger.info(max_dc_str)
            stdev_dc_str = f"Stdev degree centrality: {np.std(list(dc_values))}"
            self._results.append(stdev_dc_str)
            logger.info(stdev_dc_str)

            save_file(filename + suffix + ext, self._results)
            progressbar.update()

            # Local cluster coefficient
            logger.info("Calculating local cluster coefficients.")
            if self.config.weighted:
                lcc_values = self._local_cluster_coefficient(weight='weight')
            else:
                lcc_values = self._local_cluster_coefficient()
            avg_lcc_str = f"Average local cluster coefficient: {np.mean(list(lcc_values))}"
            self._results.append(avg_lcc_str)
            logger.info(avg_lcc_str)
            min_lcc_str = f"Minimum local cluster coefficient: {np.min(list(lcc_values))}"
            self._results.append(min_lcc_str)
            logger.info(min_lcc_str)
            max_lcc_str = f"Maximum local cluster coefficient: {np.max(list(lcc_values))}"
            self._results.append(max_lcc_str)
            logger.info(max_lcc_str)
            stdev_lcc_str = f"Stdev local cluster coefficient: {np.std(list(lcc_values))}"
            self._results.append(stdev_lcc_str)
            logger.info(stdev_lcc_str)

            save_file(filename + suffix + ext, self._results)
            progressbar.update()

            # Density
            density_str = f"Density: {self._density()}"
            self._results.append(density_str)
            logger.info(density_str)

            save_file(filename + suffix + ext, self._results)
            progressbar.update()

            # Diameter
            diameter_str = f"Diameter: {self._diameter()}"
            self._results.append(diameter_str)
            logger.info(diameter_str)

            save_file(filename + suffix + ext, self._results)
            progressbar.update()

            # Average shortest path
            avg_shortest_path_str = f"Average shortest path: {self._avg_shortest_path()}"
            self._results.append(avg_shortest_path_str)
            logger.info(avg_shortest_path_str)

            save_file(filename + suffix + ext, self._results)
            progressbar.update()

            # Global clustering coefficient
            global_cc_str = f"Global clustering coefficient: {self._global_cluster_coefficient()}"
            self._results.append(global_cc_str)
            logger.info(global_cc_str)

            save_file(filename + suffix + ext, self._results)
            progressbar.update()

            # Graph clique number
            clique_number_str = f"Graph clique number: {self._clique_number()}"
            self._results.append(clique_number_str)
            logger.info(clique_number_str)

            save_file(filename + suffix + ext, self._results)
            progressbar.update()

            # Graph number of cliques
            nr_of_cliques_str = f"Number of cliques: {self._nr_of_cliques()}"
            self._results.append(nr_of_cliques_str)
            logger.info(nr_of_cliques_str)

            save_file(filename + suffix + ext, self._results)
            progressbar.update()

    def _node_count(self):
        """Return node count.

        :return: Node count
        :rtype: `int`
        """
        return len(list(self.g.nodes))

    def _edge_count(self):
        """Return edge count.

        :return: Edge count
        :rtype: `int`
        """
        return len(list(self.g.edges))

    def _component_count(self):
        """Return component count.

        :return: Component count
        :rtype: `int`
        """
        return len(list(nx.connected_components(self.g)))

    def _max_component(self, return_component=False):
        """Return largest component size or largest component.

        :param return_component: Whether component should be returned instead of only the size
        :type return_component: `bool`
        :return: Largest component size or largest component itself
        :rtype: `set` or `int`
        """
        if return_component:
            return max(nx.connected_components(self.g), key=len)
        else:
            return len(list(max(nx.connected_components(self.g), key=len)))

    def _avg_degree(self):
        """Return average node degree.

        :return: Average node degree
        :rtype: `float`
        """
        return float(self.g.size()) / self.g.order()

    def _degree_distribution(self, weight=None):
        """Calculate, plot and save degree distributions.

        :param weight: Edge attribute that contains weight information
        :type weight: `str`
        """
        degreeview = self.g.degree(weight=weight)
        degree_sequence = sorted([d for n, d in degreeview], reverse=True)
        degree_count = collections.Counter(degree_sequence)
        deg, cnt = zip(*degree_count.items())
        deg = [int(x) for x in deg]  # convert to integer values

        plt.rc('axes', titlesize=14)
        plt.rc('axes', labelsize=14)
        plt.rc('xtick', labelsize=14)
        plt.rc('ytick', labelsize=14)

        # Relative density
        fig, ax = plt.subplots(figsize=(10, 6))
        ax.hist(degree_sequence, bins=range(0, int(max(degree_sequence))),
                density=True, histtype='bar', rwidth=0.8, color="orange")
        ax.set_yscale('log')
        ax.set_xscale('log')
        ax.set_xlabel('Knotengrad')
        ax.set_ylabel('Relative Dichte')
        filename, suffix, ext = MainUtils.get_file_info(
            self.config.img + self.config.config_id + "_degree_density.png", suffix=self.config.name
        )
        plt.savefig(filename + suffix + ext)
        if self.config.verbose:
            plt.show()
        plt.close()

        # Histogram
        fig, ax = plt.subplots(figsize=(10, 6))
        ax.bar(deg, cnt, color="orange")
        ax.set_ylabel("Anzahl an Knoten")
        ax.set_xlabel("Knotengrad")
        ax.set_xticks([d + 0.4 for d in deg])
        ax.set_xticklabels(deg)
        filename, suffix, ext = MainUtils.get_file_info(
            self.config.img + self.config.config_id + "_degree_histogram.png", suffix=self.config.name
        )
        plt.savefig(filename + suffix + ext)
        if self.config.verbose:
            plt.show()
        plt.close()

        # Degree distribution logarithmic scale
        fig, ax = plt.subplots(figsize=(10, 6))
        ax.grid(True)
        ax.set_xlabel('Knotengrad')
        ax.set_ylabel('Anzahl an Knoten')
        ax.set_xscale("log")
        ax.set_yscale("log")
        ax.plot(deg, cnt, 'o', color="orange")
        filename, suffix, ext = MainUtils.get_file_info(
            self.config.img + self.config.config_id + "_degree_distribution_log-log.png", suffix=self.config.name
        )
        plt.savefig(filename + suffix + ext)
        if self.config.verbose:
            plt.show()
        plt.close()

        # Degree distribution (normal scale)
        fig, ax = plt.subplots(figsize=(10, 6))
        ax.grid(True)
        # Count value occurrence
        _values = {}
        for d in degreeview:
            if d[1] in _values.keys():
                _values[d[1]] += 1
            else:
                _values[d[1]] = 1
        # Create histogram
        values = sorted(set(_values.keys()))
        hist = [_values[x] for x in values]
        ax.plot(values, hist, 'o', color='orange')
        ax.set_xlabel('Knotengrad')
        ax.set_ylabel('Anzahl an Knoten')
        filename, suffix, ext = MainUtils.get_file_info(
            self.config.img + self.config.config_id + "_degree_distribution.png", suffix=self.config.name
        )
        plt.savefig(filename + suffix + ext)
        if self.config.verbose:
            plt.show()
        plt.close()

    def _top_dc_nodes(self, k=5):
        """"Get top k Degree Centrality nodes.

        :param k: Number of nodes that will be collected
        :type k: `int`
        """
        top_dc = DictUtils.get_top_keys(nx.degree_centrality(self.g), 100)
        nodes = []

        count = 0
        for node_id in top_dc:
            logger.info(f"Top {count} Degree Centrality node: {self.g.degree[node_id]}, {node_id}")
            nodes.append((self.g.degree[node_id], node_id))
            count += 1
            if count == k:
                break

        self._results.append(f"Top degree centrality nodes: {nodes}")

        filename, suffix, ext = MainUtils.get_file_info(
            self.config.o, suffix=self.config.name
        )
        save_file(filename + suffix + ext, self._results)

    def _degree_centrality(self):
        """Calculate, plot and save Degree Centrality distributions.

        :return: Degree Centralities and their occurrence
        :rtype: `tuple`
        """
        sequence = sorted([round(d, 4) for n, d in nx.degree_centrality(self.g).items()], reverse=True)
        count = collections.Counter(sequence)
        dc, _ = zip(*count.items())

        plt.xlabel('Degree Centrality')
        plt.ylabel('Relative Häufigkeit')
        hist, bins = np.histogram(dc, bins=len(dc) if len(dc) < 40 else 40, density=True)
        # Normalize y axis to [0,1]
        plt.bar(bins[:-1], hist.astype(np.float32) / hist.sum(), width=(bins[1] - bins[0]) * 0.8, color='orange')
        filename, suffix, ext = MainUtils.get_file_info(
            self.config.img + self.config.config_id + "_degree_centrality.png", suffix=self.config.name
        )
        plt.savefig(filename + suffix + ext)
        if self.config.verbose:
            plt.show()
        plt.close()

        return dc

    def _local_cluster_coefficient(self, weight=None):
        """Calculate, plot and save local cluster coefficients.

        :param weight: Edge attribute that contains weight information
        :type weight: `str`
        :return: Local cluster coefficients and their occurrence
        :rtype: `tuple`
        """
        sequence = sorted([round(d, 4) for n, d in nx.algorithms.cluster.clustering(self.g, weight=weight).items()],
                          reverse=True)
        count = collections.Counter(sequence)
        cc, _ = zip(*count.items())

        plt.xlabel('Lokaler Cluster-Koeffizient')
        plt.ylabel('Relative Häufigkeit')
        hist, bins = np.histogram(cc, bins=len(cc) if len(cc) < 40 else 40, density=True)
        # Normalize y axis to [0,1]
        plt.bar(bins[:-1], hist.astype(np.float32) / hist.sum(), width=(bins[1] - bins[0]) * 0.8, color='orange')
        filename, suffix, ext = MainUtils.get_file_info(
            self.config.img + self.config.config_id + "_local_cluster_coefficient.png", suffix=self.config.name
        )
        plt.savefig(filename + suffix + ext)
        if self.config.verbose:
            plt.show()
        plt.close()

        return cc

    def _density(self):
        """Return graph density.

        :return: Graph density
        :rtype: `float`
        """
        return nx.density(self.g)

    def _diameter(self):
        """Return graph diameter.

        :return: Graph diameter
        :rtype: `int`
        """
        if nx.is_connected(self.g):
            return nx.algorithms.diameter(self.g)
        else:
            return nx.algorithms.diameter(self.g.subgraph(self._max_component(return_component=True)))

    def _avg_shortest_path(self):
        """Return average shortest path.

        :return: Average shortest path
        :rtype: `float`
        """
        if nx.is_connected(self.g):
            return nx.average_shortest_path_length(self.g, weight='weight')
        else:
            return nx.average_shortest_path_length(self.g.subgraph(self._max_component(return_component=True)),
                                                   weight='weight')

    def _global_cluster_coefficient(self):
        """Return global cluster coefficient.

        :return: Global cluster coefficient
        :rtype: `float`
        """
        return nx.average_clustering(self.g)

    def _clique_number(self):
        """Return clique number (i.e., size of largest clique).

        :return: Size of largest clique
        :rtype: `int`
        """
        return nx.graph_clique_number(self.g)

    def _nr_of_cliques(self):
        """Return number of cliques.

        :return: Number of cliques
        :rtype: `int`
        """
        return nx.graph_number_of_cliques(self.g)



# !/usr/bin/python
# -*- coding: utf-8 -*-

"""This module provides utilities to handle tar files."""

import os
import tarfile


def tar(path, name):
    """Put contents from `path` to tar.gz archive in `name`."""
    if os.path.isfile(path):
        with tarfile.open(name, "w:gz") as tar_file:
            tar_file.add(path, arcname=os.path.basename(path))
            return tar_file
    elif os.path.isdir(path):
        with tarfile.open(name, "w:gz") as tar_file:
            for root, dirs, files in os.walk(path):
                for file in files:
                    tar_file.add(os.path.join(root, file))
        return tar_file
    else:
        raise ValueError("Path is neither file nor directory!")



# !/usr/bin/python
# -*- coding: utf-8 -*-

"""This module provides basic classes used for configuration of other objects."""

from abc import ABC


class BibliometaConfiguration(ABC):
    """The abstract :class:`~bibliometa.configuration.BibliometaConfiguration` defines a configuration object
    for use as basis for Bibliometa classes. Configuration values can be accessed using dot notation (e.g.,
    self.config.name).
    """

    def __init__(self, d, **kwargs):
        """Construct a new :class:`~bibliometa.configuration.BibliometaConfiguration`.

        :param d: Dictionary with configuration key--values pairs. This dictionary will be loaded as
            initial configuration when a new instance of :class:`~bibliometa.configuration.BibliometaConfiguration`
            is constructed.
        :type d: `dict`
        :param kwargs: Arbitrary keyword arguments that are used as configuration keys and values.
            For example, `verbose=True` will make available a configuration key `verbose` with the value
            `True` (i.e., `self.config.verbose` will then return `True`). Configuration can be set during
            initialization as well as after constructing a class instance by calling the `set_config` method.
        """
        self.config = Config(d)
        self.set_config()  # initialize config id
        if kwargs:
            self.set_config(**kwargs)

    def __repr__(self):
        return str(self.config)

    def __str__(self):
        return str(self.config)

    def set_config(self, **kwargs):
        """Set configuration for key-value pairs given in kwargs.

        :return: The calling instance
        :rtype: `bibliometa.configuration.BibliometaConfiguration`
        """
        # update kwargs
        self.config.__dict__.update((k, v) for k, v in kwargs.items())
        # update config_id
        try:
            self.config.__dict__["config_id"] = f"{str(self.config.n)}_" \
                                                f"{str(self.config.e)}_" \
                                                f"{str(self.config.sim)}_" \
                                                f"{str(self.config.t)}"
        except Exception:
            pass

        return self

    def get_config(self, *args):
        """Get configuration. If no args given, the full configuration is returned.
        Otherwise, only the configuration parameters given in args are returned.

        :return: The calling instance if no args given, else a Config object
        :rtype: `bibliometa.configuration.BibliometaConfiguration` or
            `bibliometa.configuration.Config`
        """
        # update config_id
        try:
            self.config.__dict__["config_id"] = f"{str(self.config.n)}_" \
                                                f"{str(self.config.e)}_" \
                                                f"{str(self.config.sim)}_" \
                                                f"{str(self.config.t)}"
        except Exception:
            pass

        if args:
            _cfg = Config({k: self.config.__dict__[k] for k in args})
            return _cfg
        else:
            return self


class Config:
    """The :class:`~bibliometa.configuration.Config` defines a configuration object whose values can be
    accessed using dot notation (e.g., self.config.name).
    """

    def __init__(self, d):
        """Construct a new :class:`~bibliometa.configuration.Config`."""
        self.__dict__.update((k, v) for k, v in d.items())

    def __repr__(self):
        return '\n'.join(str(v) for v in self.__dict__.items())

    def __str__(self):
        return '\n'.join(str(v) for v in self.__dict__.items())



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""This file provides constants used throughout other modules."""

LOGGING_FILENAME = "./logs/{time}.log"
LOGGING_FORMAT = "{time} {level} {message}"

CSV_JSON_CONVERSION_CONFIG_DEFAULT = {
    "i": None,  # input CSV file
    "o": None,  # path to file name where JSON will be saved
    "from_": None,  # starting year
    "to": None,  # end year
    "step": 10,  # interval between two years
    "fields": [  # fields that will be extracted from the CSV file; mandatory
        {"content": ("515", "a"),
         "type": ("515", "0"),
         "categories": ["actv"]},
        {"content": ("350", "a"),
         "type": ("350", "0"),
         "categories": ["acti"]},
    ],
    "subfield_sep": "$",  # separator that splits fields and their subfields
    "split_char": " ### ",  # character used to split multiple values in a cell
    "csv_sep": "\t",  # CSV separator
    "datefield": ("340", "x"),  # field and subfield of date field used; mandatory
    "date_indicator": ["0", "1"],  # 0 == biographical | 1 == activity
    "interval_lower": 10,  # mandatory if not all data sets have a start and end year
    "interval_upper": 10,  # mandatory if not all data sets have a start and end year
    "log": None,  # path to log file
    "log_level_std": "INFO",  # log severity level on standard output
    "log_level_file": "DEBUG",  # log severity level in log file
    "verbose": False,  # if additional information is shown during program execution
    "encoding": "utf-8"  # file encoding used
    # columns in input CSV must be in the form "field + subfield separator + subfield";
    # multiple values in a cell must be delimited by "split_char";
    # each content column needs a type column where each entry in the content column
    # has a related entry in the type column;
    # only those types defined under the "categories" key are considered
}

JSON_EDGELIST_CONVERSION_CONFIG_DEFAULT = {
    "i": None,  # input JSON file; mandatory
    "o": None,  # output similarity file; mandatory
    "create_corpus": False,  # if a graph corpus will be created
    "corpus": None,  # path to graph corpus file
    "name": "",  # unique name for single conversion step (needed when script is called from a loop)
    "fields": None,  # fields and subfields to consider; mandatory
    "swap": False,  # whether keys from input file become values in output file
    "sim_functions": None,  # similarity functions
    "archive": True,  # whether output similarity file is put into an archive
    "archive_ext": ".tar.gz",  # file extension of output archive
    "csv_sep": "\t",  # CSV separator
    "log": None,  # path to log file
    "log_level_std": "INFO",  # log severity level on standard output
    "log_level_file": "DEBUG",  # log severity level in log file
    "verbose": False,  # if additional information is shown during program execution
    "encoding": "utf-8"  # file encoding used
}

GRAPH_ANALYSIS_CONFIG_DEFAULT = {
    "i": None,  # input similarity file
    "o": None,  # path where output text file with graph analysis are stored; mandatory
    "img": None,  # path where output images are stored; mandatory
    "create_graphml": False,  # if GraphML file should be created when creating graph from similarity file
    "graphml": None,  # path where GraphML input/output files are stored; mandatory
    "n": None,  # name of nodes; mandatory
    "e": None,  # name of edges; mandatory
    "sim": None,  # name of similarity function; mandatory
    "sim_functions": [],  # list of similarity functions in the same order as given in the input file; mandatory
    "weighted": False,  # if similarity functions creates weighted edges
    "t": 0,  # threshold for similarity function
    "reload": True,  # if creating graph from similarity file is enforced, even if GraphML file is already available
    "name": "",  # string that uniquely identifies a run if graph analysis is conducted in a loop
    "chunksize": 1000000,  # edges that will be loaded per chunk
    "csv_sep": "\t",  # CSV separator
    "log": None,  # path to log file
    "log_level_std": "INFO",  # log severity level on standard output
    "log_level_file": "DEBUG",  # log severity level in log file
    "verbose": False,  # if additional information is shown during program execution
    "encoding": "utf-8"  # file encoding used
}

GRAPH_VISUALISATION_MAP_CONFIG_DEFAULT = {
    "graphml": None,  # input GraphML file
    "o": None,  # output folder for image files
    "o_formats": ("pdf", "svg"),  # output formats that will be created
    "degrees": None,  # output folder where node degree information is saved
    "shapefile": "../data/shapefiles/DEU_adm1/DEU_adm1.shp",  # input shapefile
    "shapefile_color": "grey",  # shapefile color
    "degree_node_color": "black",  # color of nodes in degree images
    "coordinates": "../data/german_cities.csv",  # CSV file with coordinates information
    "coordinates_sep": ",",  # CSV separator in coordinates input file
    "geo_col": "city",  # column with city information
    "crs": "epsg:4326",  # Coordinate Reference System used
    "name": None,  # identifier for a visualization run that is appended to output images filenames; mandatory
    "figsize": (55, 55),  # size of output figures
    "log": None,  # path to log file
    "log_level_std": "INFO",  # log severity level on standard output
    "log_level_file": "DEBUG",  # log severity level in log file
    "verbose": False,  # if additional information is shown during program execution
    "encoding": "utf-8",  # file encoding used
}



# !/usr/bin/python
# -*- coding: utf-8 -*-

"""This module provides a class for similarity function definitions and similarity calculations."""

import csv
import itertools
from collections import OrderedDict

from loguru import logger
from scipy.special import binom
from tqdm.auto import tqdm

from bibliometa.utils.utils import MainUtils


class Similarity:
    """The :class:`~bibliometa.graph.similarity.Similarity` provides functions to define and calculate
    different types of similarity.
    """

    @staticmethod
    def calculate(corpus, config):
        """Calculate similarity between data sets.

        :param corpus: Graph corpus that contains the data on which similarity calculation will be based
        :type corpus: `dict`
        :param config: Configuration object
        :type config: `bibliometa.configuration.Config`
        """
        results = []
        record_count = 0

        # calculate size of progress bar
        combinations = binom(len(list(corpus.keys())), 2)
        factor = MainUtils.get_factor(combinations)
        progress_max_value = int(combinations / factor)

        # Calculate similarity for each node--node pair in the corpus
        with tqdm(total=progress_max_value) as progressbar:
            for key1, key2 in itertools.combinations(corpus, 2):
                similarity_dict = Similarity._get_similarity(
                    set(corpus[key1]), set(corpus[key2]), config.sim_functions
                )
                # consider this combination only if at least one similarity function returned a value > 0
                if any(x > 0 for x in list(similarity_dict.values())):
                    _tmp_results = [record_count, key1, key2] + [similarity_dict[s] for s in similarity_dict.keys()]
                    results.append(_tmp_results)

                # get progress
                record_count += 1
                if record_count % factor == 0:
                    progressbar.update()

            progressbar.close()

            # TODO: implement chunks for larger data sets? If yes, also implement saving temporary results/progress
            progress = 1.0
            Similarity._write_results_and_progress(results, config, progress)

    @staticmethod
    def _get_similarity(a, b, sim_functions):
        """Get similarity between a and b for different similarity functions. Get arguments that need to
        be passed to similarity functions from configuration, where necessary.

        :param a: Set of values for item a
        :type a: `set`
        :param b: Set of values for item b
        :type b: `set`
        :param sim_functions: Similarity functions
        :type sim_functions: `list` of `dict`
        :return: OrderedDict with similarity functions as keys and their values
        :rtype: `OrderedDict`
        """
        d = OrderedDict()

        # default arguments for similarity functions
        f = None  # function that calculates similarity value
        t = 0  # threshold for similarity value

        # iterate over similarity functions, get arguments and call the function
        for sim_func in sim_functions:
            if "args" in sim_func.keys():
                if "f" in sim_func["args"]:
                    f = sim_func["args"]["f"]
                if "t" in sim_func["args"]:
                    t = sim_func["args"]["t"]

            d[sim_func["name"]] = sim_func["function"](
                a, b, f, t
            )

        return d

    @staticmethod
    def _write_results_and_progress(results, config, progress):
        """Write (temporary) results and progress to the appropriate files.

        :param results: List of similarity results for node-node pairs
        :type results: `list`
        :param config: Configuration object
        :type config: `bibliometa.configuration.Config`
        :param progress: Progress values
        :type progress: `float`
        """
        filename, suffix, ext = MainUtils.get_file_info(config.o, suffix=config.name)
        with open(filename + suffix + ext, "w", newline="") as f:
            wr = csv.writer(f, delimiter=config.csv_sep)
            wr.writerows(results)
        logger.info(f"(Temporary) Results written to file {filename + suffix + ext}.")

        filename, suffix, ext = MainUtils.get_file_info(config.log, suffix=config.name)
        with open(filename + suffix + "_progress" + ext, "w", newline="") as f:
            f.write(str(progress))
        logger.info(f"Progress written to file {filename + suffix + ext}.")

    class Functions:
        """This class contains predefined similarity functions.
        """

        @staticmethod
        def mint(a, b, f, t=0):
            """a and b are considered similar if the size of their intersection is greater than or
            equal to t.

            :param a: Set of values for item a
            :type a: `set`
            :param b: Set of values for item b
            :type b: `set`
            :param f: This value (or the result of this function) will be returned if similarity between
                a and b >= t
            :type f: function or `int`
            :param t: Threshold
            :type t: `int`
            :return: Similarity value
            :rtype: `float` or `int`
            :raise ValueError: If f is neither a function nor an `int` or `float`
            """
            if callable(f):
                return f(a, b) if len(list(a.intersection(b))) >= t else 0
            elif isinstance(f, int) or isinstance(f, float):
                return f if len(list(a.intersection(b))) >= t else 0
            else:
                raise ValueError("Parameter 'f' is neither function nor int or float!")

        @staticmethod
        def jaccard(a, b, f, t=0):
            """The Jaccard Index. a and b are considered similar if the size of their intersection divided by their
            union is greater than or equal to t.

            :param a: Set of values for item a
            :type a: `set`
            :param b: Set of values for item b
            :type b: `set`
            :param f: This value (or the result of this function) will be returned if similarity between
                a and b >= t
            :type f: function or `int`
            :param t: Threshold
            :type t: `int`
            :return: Similarity value
            :rtype: `float` or `int`
            :raise ValueError: If f is neither a function nor an `int` or `float`
            """
            intersection = len(list(a.intersection(b)))
            union = (len(list(a)) + len(list(b))) - intersection
            jacc = float(intersection) / union
            if callable(f):
                return f(jacc) if union > t else 0
            elif isinstance(f, int) or isinstance(f, float):
                return jacc if union > t else 0
            else:
                raise ValueError("Parameter 'f' is neither function nor int or float!")

        @staticmethod
        def overlap(a, b, f, t=0):
            """The overlap score. a and b are considered similar if the size of their intersection divided by the
            minimum set length of a and b is greater than or equal to t.

            :param a: Set of values for item a
            :type a: `set`
            :param b: Set of values for item b
            :type b: `set`
            :param f: This value (or the result of this function) will be returned if similarity between
                a and b >= t
            :type f: function or `int`
            :param t: Threshold
            :type t: `int`
            :return: Similarity value
            :rtype: `float` or `int`
            :raise ValueError: If f is neither a function nor an `int` or `float`
            """
            intersection = len(list(a.intersection(b)))
            ovlp = float(intersection) / min(len(list(a)), len(list(b)))
            if callable(f):
                return f(ovlp) if min(len(list(a)), len(list(b))) > t else 0
            elif isinstance(f, int) or isinstance(f, float):
                return ovlp if min(len(list(a)), len(list(b))) > t else 0
            else:
                raise ValueError("Parameter 'f' is neither function nor int or float!")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['bibliometa', 'bibliometa.graph', 'bibliometa.utils']

package_data = \
{'': ['*']}

install_requires = \
['cartopy>=0.19.0,<0.20.0',
 'cython>=0.29.24,<0.30.0',
 'geopandas>=0.9.0,<0.10.0',
 'loguru>=0.5.3,<0.6.0',
 'matplotlib>=3.4.2,<4.0.0',
 'networkx>=2.6.1,<3.0.0',
 'numpy>=1.21.1,<2.0.0',
 'pandas>=1.3.0,<2.0.0',
 'scipy>=1.7.0,<2.0.0',
 'shapely>=1.7.1,<2.0.0',
 'tqdm>=4.61.2,<5.0.0']

setup_kwargs = {
    'name': 'bibliometa',
    'version': '0.1.1',
    'description': 'A package for manipulating, converting and analysing bibliographic metadata',
    'long_description': '# Bibliometa\n\n[![Documentation Status](https://readthedocs.org/projects/bibliometa/badge/?version=latest)](https://bibliometa.readthedocs.io/en/latest/?badge=latest)\n[![PyPI version](https://badge.fury.io/py/bibliometa.svg)](https://badge.fury.io/py/bibliometa)\n\nBibliometa is a python library for manipulating, converting and analysing bibliographic metadata, with a focus on graph analysis.\n\n*Repository*: https://github.com/alueschow/bibliometa\n\n*Package*: https://pypi.org/project/bibliometa/\n\n*License*: MIT\n\n-----\n\n## Installation\n* Use pip: ```pip install bibliometa```\n\n## Development\n* Clone this repository: ```git clone https://github.com/alueschow/bibliometa.git```\n* Run ```poetry install``` to install all necessary dependencies\n+ After development, run ```poetry export --without-hashes -f requirements.txt --output requirements.txt ``` to create a _requirements.txt_ file with all dependencies. This file is needed to create the documentation on [ReadTheDocs](https://readthedocs.org/).\n*  Run ```poetry run sphinx-apidoc -f -o source ../src/bibliometa``` and then ```poetry run make html``` from within the _docs_ folder to create a local documentation using [Sphinx](https://www.sphinx-doc.org/en/master/).\n',
    'author': 'Andreas Lüschow',
    'author_email': 'lueschow@sub.uni-goettingen.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://bibliometa.readthedocs.io',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""This module provides utility classes and functions whose usage is not limited to a specific context."""

import json
import os
import sys

from loguru import logger

from bibliometa.config import LOGGING_FILENAME, LOGGING_FORMAT


class MainUtils:
    """The :class:`~bibliometa.utils.utils.MainUtils` provides generic utilities.
    """

    @staticmethod
    def get_file_info(path, suffix=""):
        """Get filename, suffix and file extension from a path.

        :param path: Path to a file
        :type path: `str`
        :param suffix: Suffix that should be added to a filename (optional)
        :type suffix: `str`
        :return: Filename, suffix and file extension
        :rtype: tuple of `str`
        """
        filename, ext = os.path.splitext(path)
        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        s = MainUtils._get_suffix(suffix)

        return filename, s, ext

    @staticmethod
    def _get_suffix(s):
        """Create a filename suffix by adding an underscore before parameter `s`.

        :param s: Suffix that will be preceded by an underscore
        :type s: `str`
        :return: "_" + suffix, if suffix length > 0
        :rtype: `str`
        """
        suffix = ""
        if len(str(s)) > 0:
            suffix = "_" + str(s)
        return suffix

    @staticmethod
    def get_factor(size):
        """Calculate factor to keep max value of progress bar below 100.

        :param size: Number of total values
        :type size: `int`
        :return: Factor by which number of values needs to be divided
        :rtype: `int`
        """
        factor = 1
        while size > 100:
            size = int(size / 10)
            factor *= 10
        return factor


class DictUtils:
    """The :class:`~bibliometa.utils.utils.DictUtils` provides generic utilities."""

    # Default values
    _ENCODING = "utf-8"

    @staticmethod
    def remove_keys(i, o, k, encoding=_ENCODING):
        """Remove all keys from a given dict i (in a JSON file) if not in k and save the remaining dict as JSON in o.

        :param i: Path to input JSON file
        :type i: `str`
        :param o: Path to output JSON file
        :type o: `str`
        :param k: List of keys to be removed
        :type k: `list`
        :param encoding: File encoding
        :type encoding: `str`
        """
        # Set up logging
        logger.remove()
        logger.add(LOGGING_FILENAME, format=LOGGING_FORMAT, level="DEBUG")
        logger.add(sys.stderr, level="INFO")

        with open(i, "r", encoding=encoding) as f:
            in_data = json.load(f)

        out_data = dict()
        for key in in_data.keys():
            if key in list(k):
                out_data[key] = in_data[key]

        with open(o, 'w', encoding=encoding) as f:
            json.dump(out_data, f, indent=4)
            logger.info(f"Keys from {i} were removed (except keys in {k}). New dictionary was written to {o}.")

    @staticmethod
    def remove_empty_entries(d):
        """Remove keys from dict d that have no values.

        :param d: A dictionary
        :type d: `dict`
        :return: Input dictionary without empty entries.
        :rtype: `dict`
        """
        if isinstance(d, dict):
            return {
                k: v
                for k, v in ((k, DictUtils.remove_empty_entries(v)) for k, v in d.items())
                if v
            }
        if isinstance(d, list):
            return [v for v in map(DictUtils.remove_empty_entries, d) if v]
        return d

    @staticmethod
    def merge(a, b):
        """Merge two dictionaries.

        :param a: A dictionary
        :type a: `dict`
        :param b: Another dictionary
        :type b: `dict`
        :raise: KeyError if a key is found in both dictionaries
        :return: Dictionary a merged with b
        :rtype: `dict`
        """
        intersection = set(a.keys()).intersection(set(b.keys()))
        if len(intersection) > 0:
            raise KeyError(f"Duplicate key found: {intersection}.")
        else:
            a.update(b)
            return a

    @staticmethod
    def get_top_keys(d, k):
        """Get keys with highest values from a dictionary.

        :param d: A dictionary
        :type d: `dict`
        :param k: Top k elements that will be returned
        :type k: `int`
        :return: List of tuples (value, key) for top k keys
        :rtype: `list`
        """
        items = sorted(d.items(), reverse=True, key=lambda x: x[1])
        return map(lambda x: x[0], items[:k])

    @staticmethod
    def sort_by_key(d):
        """Sort a dictionary alphabetically by its keys.

        :param d: A dictionary
        :type d: `dict`
        :return: The sorted dictionary
        :rtype: `dict`
        """
        return dict(sorted(d.items(), key=lambda x: str(x[0])))

    @staticmethod
    def save_to_json(d, f, encoding=_ENCODING):
        """Save a dictionary to a JSON file.

        :param d: A dictionary
        :type d: `dict`
        :param f: Path to file
        :type f: `str`
        :param encoding: File encoding
        :type encoding: `str`
        :raise FileNotFoundError: If f does not point to a file
        """
        try:
            with open(f, 'w', encoding=encoding) as file:
                json.dump(d, file, indent=4)
        except FileNotFoundError as e:
            raise e

    @staticmethod
    def read_from_json(f, encoding=_ENCODING):
        """Read a dictionary from a JSON file.

        :param f: Path to file
        :type f: `str`
        :param encoding: File encoding
        :type encoding: `str`
        :return: Dictionary loaded from JSON file
        :rtype: `dict`
        :raise FileNotFoundError: If f does not point to a file
        """
        try:
            with open(f, "r", encoding=encoding) as file:
                return json.load(file)
        except FileNotFoundError as e:
            raise e
